import time
from threading import Thread

from py_hydropi.lib.metrics.reporters import reporter_factory
from py_hydropi.lib.threaded_daemon import ThreadedDaemon


class MetricCollectorController(ThreadedDaemon):
    def __init__(self, db, config, collectors: list):
        super().__init__()
        self.db = db
        self.config = config.config
        self.reporter = reporter_factory(self.config.get('reporter'))
        self.frequency = self.config.get('frequency', 10)
        self.collectors = [self._init_collector(c) for c in collectors]

    def add_collector(self, collector):
        self.collectors.append(collector(self.db, self.reporter))

    def _init_collector(self, collector):
        self.logger.debug('initializing collector: {}'.format(collector.__name__))
        return collector(self.db, self.reporter)

    def _main_loop(self):
        while self._continue:
            thread_pool = []
            for c in self.collectors:
                for id_, value in c.push_all():
                    t = Thread(target=self.reporter.push, args=[self.log_result, id_, value])
                    thread_pool.append(t)
                    t.start()
            for t in thread_pool:
                t.join()
            time.sleep(self.frequency)

    def log_result(self, res):
        self.logger.error('Metrics report returned status:{} reason:{}'.format(*res))



import os

if os.environ.get('PY_HYDROPI_TESTING', '').lower() != 'true':
    from hcsr04sensor import sensor
else:
    class Sensor:
        class Measurement:
            def __init__(self, trig, echo, temperature=20, sample_size=1):
                assert type(trig) == int
                assert type(echo) == int
                assert type(temperature) in (float, int)
                assert type(sample_size) == int

                self.trig = trig
                self.echo = echo
                self.temperature = temperature
                self.sample_size = sample_size
                self._value = 50

            def raw_distance(self, sample_size=1, **_):
                assert type(sample_size) == int
                return self._value
    sensor = Sensor()

from py_hydropi.lib.modules.inputs import Input


class UltrasonicInput(Input):
    provides = ('HC_SR04', 'HCSR04')
    frequency = 30

    def __init__(self, channels=None, pi_timer=None, correction=None, **kwargs):
        super().__init__(**kwargs)
        assert channels and pi_timer

        self.pi_timer = pi_timer
        self.echo = channels.get('echo')
        self.trigger = channels.get('trigger')
        self.channels = channels
        self._correction_identifier = correction
        if correction is not None:
            self.correction = self._get_correction_input()  # type: Input
        else:
            self.correction = None  # type: Input

    def _get_correction_input(self):
        return self.pi_timer.db.get_input(self._correction_identifier)  # type: Input

    def _read(self):
        if self.correction is None:
            correction = self._get_correction_input()
            if isinstance(correction, Input):
                self.correction = correction

        try:
            value = sensor.Measurement(
                self.trigger,
                self.echo,
                temperature=(
                    self.correction.value
                    if self.correction is not None
                    and self.correction.value is not None
                    else 20))
            distance = value.raw_distance(sample_size=11, sample_wait=0.2)
            return round(distance, 2)
        except Exception as e:
            self.logger.error('{} channels:{}'.format(e, self.channels))



import time

from py_hydropi.lib import Output, parse_simple_time_string
from py_hydropi.lib.iter_utils import JSONableDict
from py_hydropi.lib.modules.inputs import Input
from py_hydropi.lib.modules.switch import Switch
from py_hydropi.lib.modules.timer import SimpleTimer


class ThresholdSwitch(Switch):
    RISING = 1
    FALLING = -1
    INACTIVE = 0

    def __init__(self, target: float=None, upper: float=None,
                 lower: float=None, input_: Input=None, poll_sec=1,
                 alter_target=None, rpi_timer=None):

        super().__init__()
        self.rpi_timer = rpi_timer
        self._target = target
        self._upper = upper
        self._lower = lower
        self._input = input_
        self._state = self.INACTIVE
        self._rising_object = None
        self._falling_object = None
        self._rising_activated = False
        self._falling_activated = False
        self._poll_sec = poll_sec

        if alter_target:
            altering_output = alter_target.get('output')
            altering_input = alter_target.get('input')
            self.alter_when = alter_target.get('when')
            if altering_output:
                self.altering_obj = self.rpi_timer.db.get_output(altering_output)
                self.alter_func = lambda o: (o.outputs_activated or False) == (self.alter_when == 'activated')

            elif altering_input:
                self.altering_obj = self.rpi_timer.db.get_input(sensor_id=altering_input)

                if '<' in self.alter_when:
                    self.alter_func = lambda o: o.value < float(self.alter_when.split(' ')[1])

                elif '>' in self.alter_when:
                    self.alter_func = lambda o: o.value > float(self.alter_when.split(' ')[1])

            self.alter_by = alter_target.get('alter_by')

            assert self.altering_obj
            assert self.alter_by
            assert self.alter_func

        self.threshold_timer = None

    @property
    def all_outputs(self):
        out = super().all_outputs
        out.extend([self._rising_object, self._falling_object])
        return out

    def to_json(self):
        return JSONableDict(target=self._target,
                            poll_sec=self._poll_sec,
                            upper_limit=self._upper,
                            lower_limit=self._lower)

    @staticmethod
    def factory(**kwargs):
        if kwargs.get('periodic'):
            kwargs = {k: v for k, v in kwargs.items() if k in ('target', 'upper', 'lower', 'input_',
                                                               'poll_sec', 'periodic', 'rpi_timer',
                                                               'alter_target')}
            return PeriodicThresholdSwitch(**kwargs)
        elif kwargs.get('min_duty_cycle'):
            kwargs = {k: v for k, v in kwargs.items() if k in ('target', 'upper', 'lower', 'input_',
                                                               'poll_sec', 'min_duty_cycle',
                                                               'rpi_timer', 'alter_target')}
            return TargetThresholdSwitch(**kwargs)

    @classmethod
    def load_config(cls, raspberry_pi_timer, config):
        objects = {group: cls.factory(
            target=group_settings.get('target'),
            upper=group_settings.get('upper').get('limit'),
            lower=group_settings.get('lower').get('limit'),
            min_duty_cycle=group_settings.get('min_duty_cycle'),
            input_=raspberry_pi_timer.db.get_input(sensor_id=group_settings.get('input')),
            periodic=group_settings.get('periodic'),
            alter_target=group_settings.get('alter_target'),
            rpi_timer=raspberry_pi_timer
        ).set_rising_object(
            Output(pi_timer=raspberry_pi_timer,
                   channel=group_settings.get('lower').get('channel'))
        ).set_falling_object(
            Output(pi_timer=raspberry_pi_timer,
                   channel=group_settings.get('upper').get('channel'))
        ) for group, group_settings in config.items()}
        return objects

    def set_rising_object(self, obj):
        self._rising_object = obj
        return self

    def set_falling_object(self, obj):
        self._falling_object = obj
        self.attached_outputs.append(obj)
        return self

    def _activate_rising_objects(self):
        if not self._rising_activated:
            self._rising_object.activate()
            self._rising_activated = True
            self._state = self.RISING

    def _activate_falling_objects(self):
        if not self._falling_activated:
            self._falling_object.activate()
            self._falling_activated = True
            self._state = self.FALLING

    def _deactivate_rising_objects(self):
        if self._rising_activated:
            self._rising_object.deactivate()
            self._rising_activated = False
        self._state = self.INACTIVE

    def _deactivate_falling_objects(self):
        if self._falling_activated:
            self._falling_object.deactivate()
            self._falling_activated = False
        self._state = self.INACTIVE

    def stop(self):
        if self.threshold_timer:
            self.threshold_timer.stop()
        super().stop()

    def _main_loop(self):
        while self._continue:
            if self._input.value:
                if hasattr(self, 'alter_by'):
                    self.logger.debug('pre-alter >{} <{} ={}'.format(self._upper, self._lower, self._target))
                    upper, lower, target = self._upper, self._lower, self._target
                    try:
                        if self.alter_func(self.altering_obj):
                            self._upper += self.alter_by
                            self._lower += self.alter_by
                            self._target += self.alter_by
                            self.logger.debug('post-alter >{} <{} ={}'.format(self._upper, self._lower, self._target))
                    except TypeError:
                        pass

                    self._sub_loop()
                    self._upper, self._lower, self._target = upper, lower, target
                    self.logger.debug('post-loop >{} <{} ={}'.format(self._upper, self._lower, self._target))
                else:
                    self._sub_loop()

            time.sleep(self._poll_sec)

    def _sub_loop(self):
        raise NotImplementedError


class PeriodicThresholdSwitch(ThresholdSwitch):
    def __init__(self, target=None, upper: float=None, lower: float=None, input_: Input=None,
                 poll_sec=1, periodic=None, rpi_timer=None, alter_target=None):

        super().__init__(target=target, upper=upper, lower=lower, input_=input_,
                         poll_sec=poll_sec, alter_target=alter_target, rpi_timer=rpi_timer)
        self._periodic = periodic
        self._periodic_active_sec = parse_simple_time_string(periodic.get('active'))
        self._poll_sec = parse_simple_time_string(periodic.get('inactive'))

    def _sub_loop(self):
        if self._input.value < self._lower:
            self._activate_rising_objects()
            time.sleep(self._periodic_active_sec)
            self._deactivate_rising_objects()

        if self._input.value > self._upper:
            self._activate_falling_objects()
            time.sleep(self._periodic_active_sec)
            self._deactivate_falling_objects()


class TargetThresholdSwitch(ThresholdSwitch):
    def __init__(self, target: float=None, upper: float=None,
                 lower: float=None, min_duty_cycle: float=0, input_: Input=None, poll_sec=1,
                 alter_target=None, rpi_timer=None):

        super().__init__(target=target, upper=upper, lower=lower, input_=input_,
                         poll_sec=poll_sec, alter_target=alter_target, rpi_timer=rpi_timer)
        self.rpi_timer = rpi_timer
        self._min_duty_cycle = min_duty_cycle

    def _sub_loop(self):
        if self._state == self.FALLING:
            if self._input.value <= self._target:
                self._deactivate_falling_objects()
        else:
            if self._min_duty_cycle and not self.threshold_timer:
                self.threshold_timer = SimpleTimer(
                    '{}s'.format(int(60 * self._min_duty_cycle)),
                    '{}s'.format(int(60 - (60 * self._min_duty_cycle))))
                self.threshold_timer.attached_outputs.append(self._falling_object)
                self.threshold_timer.start()

            if self._state == self.INACTIVE:
                if self._input.value > self._upper:
                    if self._min_duty_cycle and self.threshold_timer:
                        self.threshold_timer.stop()
                        self.threshold_timer = None

                    self._activate_falling_objects()

                elif self._input.value < self._lower:
                    self._activate_rising_objects()

            elif self._state == self.RISING:
                if self._input.value >= self._target:
                    self._deactivate_rising_objects()



def avg(list_):
    list_ = [i for i in list_ if i is not None] or [0]
    return sum(list_) / len(list_)


class JSONableDict(dict):
    def to_json(self):
        return {
            k: v.to_json() if hasattr(v, 'to_json') else v for k, v in self.items()
        }


class JSONableList(list):
    def to_json(self):
        return [v.to_json() if hasattr(v, 'to_json') else v for v in self]

    def append_if_not_in(self, o):
        if o not in self:
            self.append(o)

    def extend_if_not_in(self, iterable):
        for o in iterable:
            self.append_if_not_in(o)



import logging

import os
from systemd import journal


# noinspection PyUnusedLocal
class Logger(object):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR

    def __init__(self, obj):
        if os.environ.get('PY_HYDROPI_TESTING', '').lower() == 'true':
            class ScreenLogger(logging.Logger):
                def error(self, msg, *args, **kwargs):
                    if self.level <= Logger.ERROR:
                        print(msg.format(*args, **kwargs))

                def warning(self, msg, *args, **kwargs):
                    if self.level <= Logger.WARNING:
                        print(msg.format(*args, **kwargs))

                def debug(self, msg, *args, **kwargs):
                    if self.level <= Logger.DEBUG:
                        print(msg.format(*args, **kwargs))

                def info(self, msg, *args, **kwargs):
                    if self.level <= Logger.INFO:
                        print(msg.format(*args, **kwargs))

            self._logger = ScreenLogger(obj)
        else:
            self._logger = logging.getLogger(obj)
            if not self._logger.handlers:
                self._logger.addHandler(journal.JournaldLogHandler())

        env_level = os.environ.get('PY_HYDROPI_LOGGING', '').upper() or 'INFO'
        level = getattr(self, env_level) or self.INFO
        self._logger.setLevel(level)

        self.calling_obj = obj

    def warn(self, msg, *args, **kwargs):
        self._logger.warning('{}: WARNING: {}'.format(self.calling_obj, msg))

    def info(self, msg, *args, **kwargs):
        self._logger.info('{}: INFO: {}'.format(self.calling_obj, msg))

    def error(self, msg, *args, **kwargs):
        self._logger.error('{}: ERROR: {}'.format(self.calling_obj, msg))

    def debug(self, msg, *args, **kwargs):
        self._logger.debug('{}: DEBUG: {}'.format(self.calling_obj, msg))

    def log(self, level, msg, *args, **kwargs):
        if level == self.INFO:
            self.info(msg)

        if level == self.DEBUG:
            self.debug(msg)

        if level == self.WARNING:
            self.warn(msg)

        if level == self.ERROR:
            self.error(msg)



class SensorMetricCollector:
    def __init__(self, db, reporter):
        super().__init__()
        self.db = db
        self.reporter = reporter

    def push_all(self):
        return [(id, sensor.value) for id, sensor in self.db._inputs.items()]



from py_hydropi.lib.threaded_daemon import ThreadedDaemon


class Switch(ThreadedDaemon):
    def __init__(self):
        super().__init__()
        self.attached_outputs = []
        self.outputs_activated = None

    @property
    def all_outputs(self):
        return list(self.attached_outputs)

    @classmethod
    def load_config(cls, raspberry_pi_timer, config):
        raise NotImplementedError

    def attach_object(self, obj):
        if type(obj) in (list, tuple):
            self.attached_outputs.extend(obj)
        else:
            self.attached_outputs.append(obj)
        return self

    def _activate_objects(self):
        if not self.outputs_activated:
            for output in self.attached_outputs:
                if output.manual_control:
                    self.logger.info('output {} is manually controlled, skipping signal'.format(output.channel))
                else:
                    self.logger.info('signalling output {} to activate'.format(output.channel))
                    output.activate()
            self.outputs_activated = True

    def _deactivate_objects(self):
        if self.outputs_activated:
            for output in self.attached_outputs:
                if output.manual_control:
                    self.logger.info('output {} is manually controlled, skipping signal'.format(output.channel))
                else:
                    self.logger.info('signalling output {} to deactivate'.format(output.channel))
                    output.deactivate()
            self.outputs_activated = False

    def stop(self):
        super().stop()
        self._deactivate_objects()
